
subroutine PHYrad_top(Dis_ST)
    ! +------------------------------------------------------------------------+
    ! | MAR PHYSICS (INSOL)                                    15-11-2007  MAR |
    ! |   subroutine PHYrad_top computes                                       |
    ! |      Time Insolation at the Top of the Atmosphere                      |
    ! |                                                                        |
    ! +------------------------------------------------------------------------+
    ! |                                                                        |
    ! |   REFER.:   Ch.  Tricot, personal communication                        |
    ! |   ^^^^^^^   M.F. Loutre, personal communication and thesis (1993)      |
    ! |                                                                        |
    ! |   INPUT :   mmarGE, jdarGE: Month and Day of the Year                  |
    ! |   ^^^^^^^   jhurGE, minuGE, jsecGE: Hour, Minute, and Second           |
    ! |             GElat0, GElon0: Latitude, Longitude                        |
    ! |             GElatr(mx,my) : Latitude                         (radians) |
    ! |             GElonh(mx,my) : Longitude                          (hours) |
    ! |             itizGE(mx,my) : Time Zone                                  |
    ! |                                                                        |
    ! |   OUTPUT:   rsunGE        : Insolation normal to Atmosphere Top (W/m2) |
    ! |   ^^^^^^^   czenGE(mx,my) : Cosinus of the Zenithal Distance           |
    ! |                                                                        |
    ! +------------------------------------------------------------------------+

    use marctr
    use marphy
    use mardim
    use margrd
    use mar_ge
    use marsnd
    use mar_wk
    use mar_io

    implicit none

    ! +--Global Variables
    ! +  ================

    real Dis_ST

    ! +--LOCAL VARIABLES
    ! +  ===============

    integer i, j, k, m
    real pirr, xl, so
    real xllp, xee, xse
    real xlam, dlamm, anm, ranm, ranv, anv, tls
    real Tyear, step, rlam, sd, cd, deltar, delta

    real ddt, arg, et, argc, ahc
    real c1, c2, c3, s1, s2, s3
    real timdl, timh, ahor, ahorr, RadLat, chor, zenitr
    integer nj, lhc

    real omesun

    real om, ecc, perh, xob

    ! +--DATA
    ! +  ====

    data om/0.0172142d0/

    ! +--Present Day Insolation
    ! +  ----------------------
    ! Eccentricity
    data ecc/0.01673/
    ! Longitude of the Perihelion (degrees)
    data perh/102.4/
    ! Obliquity (degrees)
    data xob/23.445/

    ! +--Insolation at the Top of the Atmosphere (TIME       PARAMETERS)
    ! +  ===============================================================

    ! +--Solar declination : delta
    ! +  -------------------------

    nj = jdarGE + njyrGE(mmarGE)
    Tyear = 365.25d0
    step = 360.0d0 / Tyear

    pirr = degrad / 3600.0
    xl = perh + 180.0
    so = sin(xob * degrad)
    ! +...so    : sinus of obliquity

    xllp = xl * degrad
    xee = ecc * ecc
    xse = sqrt(1.0d0 - xee)
    xlam = (ecc / 2.0 + ecc * xee / 8.0d0) * (1.0 + xse) * sin(xllp) - xee / 4.0 * &
           (0.5 + xse) * sin(2.0 * xllp) + ecc * xee / 8.0 * (1.0 / 3.0 + xse) * &
           sin(3.0 * xllp)
    xlam = 2.0d0 * xlam / degrad
    dlamm = xlam + (nj - 80) * step
    ! +...xlam  : true long. sun for mean long. = 0
    ! +...dlamm : mean long. sun for ma-ja
    anm = dlamm - xl
    ranm = anm * degrad
    xee = xee * ecc
    ranv = ranm + (2.0 * ecc - xee / 4.0) * sin(ranm) + 5.0 / 4.0 * ecc * ecc * &
           sin(2.0 * ranm) + 13.0 / 12.0 * xee * sin(3.0 * ranm)
    anv = ranv / degrad
    tls = anv + xl
    rlam = tls * degrad
    ! +...tls   : longitude vraie   (degrees)
    ! +...rlam  : longitude vraie   (radian)
    ! +...anv   : anomalie vraie    (degrees)
    ! +...ranv  : anomalie vraie    (radian)

    sd = so * sin(rlam)
    cd = sqrt(1.0d0 - sd * sd)
    ! +...sd and cd: cosinus and sinus of solar declination angle (delta)
    ! +...sinus delta = sin (obl)*sin(lambda) with lambda = real longitude
    ! +...(Phd. thesis of Marie-France Loutre, ASTR-UCL, Belgium, 1993)

    deltar = atan(sd / cd)
    delta = deltar / degrad
    ! +...delta: Solar Declination (degrees, angle sun at equator)

    ! +--Eccentricity Effect
    ! +  -------------------

    Dis_ST = (1.0 - ecc * ecc) / (1.0 + ecc * cos(ranv))
    ddt = 1.0 / Dis_ST
    ! +...ddt  :    1   /  normalized earth's sun distance

    ! +--Insolation normal to the atmosphere (W/m2)
    ! +  ------------------------------------------

    !XF
    rsunGE = ddt * ddt * 1360.8d0

    ! +--Time Equation (Should maybe be modified in case other than present
    ! +  -------------  conditions are used, minor impact)

    arg = om * nj
    c1 = cos(arg)
    c2 = cos(2.d0 * arg)
    c3 = cos(3.d0 * arg)
    s1 = sin(arg)
    s2 = sin(2.d0 * arg)
    s3 = sin(3.d0 * arg)

    et = 0.0072d0 * c1 - 0.0528d0 * c2 - 0.0012d0 * c3 &
         - 0.1229d0 * s1 - 0.1565d0 * s2 - 0.0041d0 * s3
    ! +...et (hour)
    ! +     = difference between true solar and mean solar hour angles.
    ! +      (connected to the earth orbital rotation speed)

    ! +--Insolation at the Top of the Troposphere (Auxiliary Variables)
    ! +  ==============================================================

    ! +--Day Length, Time Sunrise and Sunset at Sounding Grid Point (iSND, jSND)
    ! +  -----------------------------------------------------------------------

    i = iSND
    j = jSND

    argc = -tan(GElatr(i, j)) * tan(deltar)
    if(abs(argc) > 1.d0) then
        ahc = 0.d0
        if(argc > 1.d0) then
            lhc = -1
            timdl = 00.d0
            ! +...   Polar  Night
        else
            lhc = 1
            timdl = 24.d0
            ! +...   Midnight Sun
        endif
        tlsrGE = 00.d0
        tlssGE = 00.d0
    else
        ahc = acos(argc)
        lhc = 0

        if(ahc < 0.d0) ahc = -ahc
        ahc = ahc / hourad
        timdl = ahc * 2.d0
        tlsrGE = 12.d0 - ahc + itizGE(i, j) - et - GElonh(i, j)
        tlssGE = tlsrGE + timdl
    endif

    tl__GE = jhurGE + minuGE / 60.d0 - itizGE(i, j)

    ! +--Time Angle
    ! +  ----------

    do j = 1, my
        do i = 1, mx
            timh = jhurGE + minuGE / 60.d0
            ahor = timh + GElonh(i, j) - 12.d0 - et
            ! +...   ahor  : time angle (hours)

            ahorr = ahor * hourad
            ! +...   ahorr : time angle (radians)

            chor = cos(ahorr)

            ! +--Solar Zenithal Distance zenitr (radians) and
            ! +  Insolation (W/m2) at the Atmosphere Top  ===
            ! +  =======================================

            czenGE(i, j) = slatGE(i, j) * sd &
                           + clatGE(i, j) * cd * chor
            czenGE(i, j) = max(czenGE(i, j), zero)

            cverGE(i, j) = czenGE(i, j)

            ! +--Slope Impact
            ! +  ------------

        enddo
    enddo

    if((jmmMAR == 0 .and. jssMAR == 0 .and. &
        ((IO_loc >= 2 .and. jhurGE == 0) .or. &
         (IO_loc >= 2 .and. mod(jhurGE, 3) == 0) .or. &
         (IO_loc >= 3))) .or. &
       IO_loc >= 7) then

        ahor = timh + GElonh(iSND, jSND) - 12.d0 - et
        zenitr = acos(czenGE(iSND, jSND)) / degrad

        write(4, 1) GElat0, GElon0, jdarGE, mmarGE, jhurGE, minuGE, jsecGE
1       format(/, ' lat.=', f6.1, 3x, 'long.=', f7.1, 4x, 'date :', i3, '-', i2, &
                ' / ', i2, ' h.UT', i3, ' min.', i3, ' sec.')
        write(4, 2) iSND, jSND, GElatr(iSND, jSND) / degrad, GElonh(iSND, jSND)
2       format(' Sounding at (', i3, i3, ') / (', f6.2, 'dg,', f6.2, 'ho)')
        write(4, 3) rsunGE * cverGE(iSND, jSND), ahor, zenitr &
            , delta
3       format(' Insolation [W/m2]  = ', f7.2, '   Hor.Angle = ', f7.2, &
               '   Zenith.Angle = ', f7.2 &
               , /, ' Solar Declination  = ', f7.2)

        if(lhc == -1) &
            write(4, 4) tlsrGE, timdl, tlssGE
4       format(' Sun Rise Time [h]  = ', f7.2, '   Day Leng. = ', f7.2, &
               '   Sun Set Time = ', f7.2, '  -- POLAR  NIGHT --')
        if(lhc == 0) &
            write(4, 5) tlsrGE, timdl, tlssGE
5       format(' Sun Rise Time [h]  = ', f7.2, '   Day Leng. = ', f7.2, &
               '   Sun Set Time = ', f7.2, '  -- SOLAR  TIME  --')
        if(lhc == 1) &
            write(4, 6) tlsrGE, timdl, tlssGE
6       format(' Sun Rise Time [h]  = ', f7.2, '   Day Leng. = ', f7.2, &
               '   Sun Set Time = ', f7.2, '  -- MIDNIGHT SUN --')
    endif

    ! +--Work Arrays Reset
    ! +  =================

    return
end
