
subroutine SISVATeSBL
    !--------------------------------------------------------------------------+
    !   MAR          SISVATeSBL                           Tue 12-JUL-2019  MAR |
    !     subroutine SISVATeSBL generates Surface Boundary Layers Properties   |
    !     (and computes usthSV since 24-sep 2018)                              |
    !--------------------------------------------------------------------------+
    !                                                                          |
    !     PARAMETERS:  klonv: Total Number of columns                          |
    !     ^^^^^^^^^^        = Total Number of continental     grid boxes       |
    !                       X       Number of Mosaic Cell per grid box         |
    !                                                                          |
    !     INPUT:   za__SV   : Surface Boundary Layer (SBL) Height          [m] |
    !     ^^^^^    VV__SV   :(SBL Top)   Wind Velocity                   [m/s] |
    !              TaT_SV   : SBL Top    Temperature                       [K] |
    !              ExnrSV   : Exner      Potential                         [-] |
    !              qsnoSV   : SBL Mean   Snow      Content             [kg/kg] |
    !              uqs_SV   : Specific   Humidity  Turbulent Flux        [m/s] |
    !              usthSV   : Threshd. friction velocity for snow erosion[m/s] |
    !              Z0m_SV   : Momentum     Roughness Length                [m] |
    !              Z0h_SV   : Heat         Roughness Length                [m] |
    !              Tsrfsv   : Surface    Temperature                       [K] |
    !              sqrCm0   : Contribution of Z0m to Neutral Drag Coefficient  |
    !              sqrCh0   : Contribution of Z0h to Neutral Drag Coefficient  |
    !                                                                          |
    !     INPUT /  LMO_SV   : Monin-Obukhov       Scale                    [m] |
    !     OUTPUT:  us__SV   : Friction  Velocity                         [m/s] |
    !     ^^^^^^   uts_SV   : Temperature         Turbulent Flux       [K.m/s] |
    !              uss_SV   : Blowing Snow        Turbulent Flux         [m/s] |
    !                                                                          |
    !     OUTPUT:  hSalSV   : Saltating Layer Height                       [m] |
    !     ^^^^^^   qSalSV   : Saltating Snow  Concentration            [kg/kg] |
    !              ram_sv   : Aerodynamic Resistance for Momentum        [s/m] |
    !              rah_sv   : Aerodynamic Resistance for Heat            [s/m] |
    !                                                                          |
    !   # OPTIONS: #BS: Blowing Snow turbulent Fluxes are computed             |
    !   # ^^^^^^^  #ss: Additional Output                                      |
    !                                                                          |
    !--------------------------------------------------------------------------+

    use marphy
    use mar_sv
    use mardsv
    use marxsv
    use marysv

    implicit none

    !  V,  dT(a-s)    Time Moving Averages
    !  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    integer ntaver, nt
    parameter(ntaver=4)
    real V__mem(klonv, ntaver)
    real VVmmem(klonv)
    common / SVeSBLmem / V__mem, VVmmem
    real T__mem(klonv, ntaver)
    real dTmmem(klonv)
    common / STeSBLmem / T__mem, dTmmem

    !$OMP threadprivate(/SVeSBLmem/,/STeSBLmem/)

    !  Internal Variables
    !  ==================
    integer ikl, icount, isn
    real VVaSBL(klonv), VVa_OK         ! VVaSBL, VVa_OK : effective SBL wind speed
    real dTa_Ts(klonv)                 ! dTa_Ts : effective SBL Temperature diff.
    real Theta0                        ! Theta0 : Potential Reference Temperature
    real LMOmom(klonv)                 ! LMOmom : Monin-Obukhov Scale Momentum
    real LMOsgn                        ! LMOsgn : Monin-Obukhov Scale Sign
    real LMOabs                        ! LMOabs : Monin-Obukhov Scale Abs.Value
    real uustar, thstar, qqstar, ssstar, thstarv, thstars, thstara
    real zetam, zetah, zeta_S, zeta_A, zeta0m, zeta0h
    real psim_s, xpsimi, psim_i, psim_z
    real psis_s, psis_z, psis_0
    real psih_s, xpsihi, psih_i, psih_z
    real psim_0, psih_0
    real CDm(klonv)                    ! CDm : Drag Coefficient, Momentum
    real CDs(klonv), rCDs(klonv)       ! CDs, rCDs : Drag Coefficient, Blown **
    real CDh(klonv)                    ! CDh : Drag Coefficient, Scalar
    real dustar, u0star, uTstar, usstar
    real sss__F, sss__N, usuth0
    real zetMAX
    real coef_m, coef_h, stab_s
    real Richar(klonv)                 ! Richar : Richardson Number

    real fac_Ri, vuzvun, Kz_vun
    character * 3 qsalt_param              ! qsalt_param : Switch for saltation flux param.
    character * 3 usth_param               ! usth_param : Switch for u*t param

    !  Internal DATA
    !  =============

    ! Theta0 :  Potential Reference Temperature
    data Theta0/288.0/

    ! zetMAX :  Strong Stability Limit
    data zetMAX/4.28/
    ! coef_m :  Stabil.Funct.for Moment.: unstab.coef.
    ! (King    et al. 1996, JGR 101(7) p.19121)
    data coef_m/20./
    ! coef_h :  Stabil.Funct.for Heat:    unstab.coef.
    data coef_h/15./

    !  Effective SBL variables
    !  =======================

    do ikl = 1, klonv
        VVaSBL(ikl) = VV__SV(ikl)
        VVaSBL(ikl) = VVmmem(ikl)
        dTa_Ts(ikl) = TaT_SV(ikl) - Tsrfsv(ikl)
        dTa_Ts(ikl) = dTmmem(ikl)
    enddo

    !  Convergence Criterion
    !  =====================

    icount = 0

1   continue
    icount = icount + 1
    dustar = 0.

    do ikl = 1, klonv

        u0star = us__SV(ikl)

        !  Turbulent Scales from previous Time Step
        !  ----------------------------------------

        u0star = max(epsi, u0star)      ! Friction Velocity     u*
        uustar = u0star * u0star       ! Friction Velocity^2  uu*
        thstar = uts_SV(ikl) / u0star       ! Temperature       theta*
        qqstar = uqs_SV(ikl) / u0star       ! Specific Humidity    qq*
        ssstar = uss_SV(ikl) / u0star       ! Blown    Snow        ss*

        !  Monin-Obukhov Stability Parameter for Momentum
        !  ----------------------------------------------

        !  Pseudo Virtual Temperature Turbulent Scale thetav*
        !  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        thstarv = thstar + Theta0 * (0.608 * qqstar) &
                  / (1.+0.608 * QaT_SV(ikl) - qsnoSV(ikl))
        thstars = sign(unun, thstarv)
        thstara = abs(thstarv)
        thstarv = max(epsi, thstara) * thstars

        !  Pseudo Obukhov Length Scale        (Gall?e et al., 2001 BLM 99, (A2) p.17)
        !  Full   Obukhov Length Scale        (when Blowing * is ##NOT## switched ON)
        !  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        LMO_SV(ikl) = Theta0 * max(epsi, uustar) &
                      / (vonkar * gravit * thstarv)

        zetah = za__SV(ikl) / LMO_SV(ikl)
        zetam = min(zetMAX, zetah)! Strong Stability Limit
        !                                                  !(Mahalov et al. 2004
        !                                                  ! GRL 31 2004GL021055)
        LMOmom(ikl) = za__SV(ikl) / (max(epsi, abs(zetam)) &
                                     * sign(unun, zetam))
        zeta0m = Z0m_SV(ikl) / LMOmom(ikl)
        zeta0h = Z0h_SV(ikl) / LMO_SV(ikl)

        !  Momentum Pseudo Stability Function (Gall?e et al. 2001, BLM 99, (11) p. 7)
        !  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        stab_s = max(zero, sign(unun, zetam))

        psim_s = -A_Turb * zetam
        xpsimi = sqrt(sqrt(unun - coef_m * min(zero, zetam)))
        psim_i = 2.*log(demi * (unun + xpsimi)) &
                 + log(demi * (unun + xpsimi * xpsimi)) &
                 - 2.*atan(xpsimi) + demi * pi
        psim_z = stab_s * psim_s + (1.-stab_s) * psim_i

        psim_s = -A_Turb * zeta0m
        xpsimi = sqrt(sqrt(unun - coef_m * min(zero, zeta0m)))
        psim_i = 2.*log(demi * (unun + xpsimi)) &
                 + log(demi * (unun + xpsimi * xpsimi)) &
                 - 2.*atan(xpsimi) + demi * pi
        psim_0 = stab_s * psim_s + (1.-stab_s) * psim_i

        !  Momentum            Turbulent Scale  u*
        !  ---------------------------------------

        !  Momentum            Turbulent Scale  u*          in case of NO Blow. Snow
        !  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        VVa_OK = max(0.000001, VVaSBL(ikl))
        sss__N = vonkar * VVa_OK
        sss__F = (sqrCm0(ikl) - psim_z + psim_0)
        usuth0 = sss__N / sss__F                ! u* if NO Blow. Snow

        us__SV(ikl) = usuth0

6001    format(18x, 9x, 'LMO  [m]=', f9.1, '   zetah[-] =', f9.3)

        !  Turbulent Scales
        !  ----------------

        !  Momentum Stability Function (Gall?e et al., 2001, BLM 99, (11) p. 7)
        !  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        stab_s = max(zero, sign(unun, zetam))

        psim_s = -A_Turb * zetam
        xpsimi = sqrt(sqrt(unun - coef_m * min(zero, zetam)))
        psim_i = 2.*log(demi * (unun + xpsimi)) &
                 + log(demi * (unun + xpsimi * xpsimi)) &
                 - 2.*atan(xpsimi) + demi * pi
        psim_z = stab_s * psim_s + (1.-stab_s) * psim_i

        psim_s = -A_Turb * zeta0m
        xpsimi = sqrt(sqrt(unun - coef_m * min(zero, zeta0m)))
        psim_i = 2.*log(demi * (unun + xpsimi)) &
                 + log(demi * (unun + xpsimi * xpsimi)) &
                 - 2.*atan(xpsimi) + demi * pi
        psim_0 = stab_s * psim_s + (1.-stab_s) * psim_i

        !  Heat     Stability Function (Gall?e et al., 2001, BLM 99, (11) p. 7)
        !  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        stab_s = max(zero, sign(unun, zetah))

        psih_s = -AhTurb * zetah
        xpsihi = sqrt(sqrt(unun - coef_h * min(zero, zetah)))
        psih_i = 2.*log(demi * (unun + xpsihi))
        psih_z = stab_s * psih_s + (1.-stab_s) * psih_i

        psih_s = -AhTurb * zeta0h
        xpsihi = sqrt(sqrt(unun - coef_h * min(zero, zeta0h)))
        psih_i = 2.*log(demi * (unun + xpsihi))
        psih_0 = stab_s * psih_s + (1.-stab_s) * psih_i

        !  Square Roots of the Drag Coefficients
        !  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        rCDhSV(ikl) = vonkar * (ExnrSV(ikl) / pcap) &
                      / (sqrCh0(ikl) - psih_z + psih_0)
        rCDmSV(ikl) = vonkar / (sqrCm0(ikl) - psim_z + psim_0)

        !  Drag Coefficients
        !  ~~~~~~~~~~~~~~~~~
        CDh(ikl) = rCDmSV(ikl) * rCDhSV(ikl)
        CDm(ikl) = rCDmSV(ikl) * rCDmSV(ikl)

        !  real Temperature Turbulent Scale theta*
        !  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        thstar = rCDhSV(ikl) * dTa_Ts(ikl) &
                 * (pcap / ExnrSV(ikl))
        uts_SV(ikl) = us__SV(ikl) * thstar

        !  Convergence Criterion
        !  =====================

        dustar = max(dustar, abs(us__SV(ikl) - u0star))

    enddo

    !     if (dustar.gt.0.0001.and.icount.lt. 6)                     go to 1

    !  Aerodynamic Resistances
    !  -----------------------

    do ikl = 1, klonv
        ram_sv(ikl) = 1./(CDm(ikl) * max(VVaSBL(ikl), epsi))
        rah_sv(ikl) = 1./(CDh(ikl) * max(VVaSBL(ikl), epsi))
    enddo

    return
endsubroutine SISVATeSBL
